<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 2/10/18
 * Time: 3:14 PM
 *
 * See: doc/THUMBNAIL.md
 */

namespace qfq;

require_once(__DIR__ . '/../helper/Support.php');
require_once(__DIR__ . '/../helper/Token.php');
require_once(__DIR__ . '/../helper/OnString.php');

//use qfq;

class Thumbnail {

    /**
     * @var Store
     */
    private $store = null;

    private $inkscape = '';
    private $convert = '';
    private $thumbnailDirSecure = '';
    private $thumbnailDirPublic = '';

    /**
     * @param bool|false $phpUnit
     */
    public function __construct($phpUnit = false) {
        $this->store = Store::getInstance();

        $this->inkscape = $this->store->getVar(SYSTEM_CMD_INKSCAPE, STORE_SYSTEM);
        $this->convert = $this->store->getVar(SYSTEM_CMD_CONVERT, STORE_SYSTEM);
        $this->thumbnailDirSecure = $this->store->getVar(SYSTEM_THUMBNAIL_DIR_SECURE, STORE_SYSTEM);
        $this->thumbnailDirPublic = $this->store->getVar(SYSTEM_THUMBNAIL_DIR_PUBLIC, STORE_SYSTEM);
    }

    /**
     * Renders a thumbnail based on $str.
     * $str: "T:<pathfilename source>|W:<dimension>|s:<0|1>"
     *
     * Argument 'T' is mandatory.
     * Argument 'W' is optional. Defaults to 'W:150x'.
     * Argument 's' is optional. Defaults to 's:1'
     *
     * @param string $str
     * @return string
     *
     * @throws UserFormException
     * @throws UserReportException
     */
    public function getImageTag($str) {

        $control = Token::explodeTokenString($str);
        Support::setIfNotSet($control, TOKEN_SIP);
        $control[TOKEN_SIP] = Token::checkForEmptyValue(TOKEN_SIP, $control[TOKEN_SIP]);

        if (empty($control[TOKEN_THUMBNAIL])) {
            throw new UserReportException("Thumbnail: source file not specified", ERROR_MISSING_REQUIRED_PARAMETER);
        }

        if (empty($control[TOKEN_THUMBNAIL_DIMENSION])) {
            $control[TOKEN_THUMBNAIL_DIMENSION] = THUMBNAIL_WIDTH_DEFAULT;
        }

        $pathFilenameSource = $control[TOKEN_THUMBNAIL];

        $dir = ($control[TOKEN_SIP] == "1") ? $this->thumbnailDirSecure : $this->thumbnailDirPublic;
        $pathFilenameThumbnail = Support::joinPath($dir, md5($pathFilenameSource . $control[TOKEN_THUMBNAIL_DIMENSION]) . '.png');

        $pathFilenameThumbnail = $this->doThumbnail($pathFilenameSource, $pathFilenameThumbnail, $control[TOKEN_THUMBNAIL_DIMENSION]);

        return $this->buildImageTag($pathFilenameThumbnail, $control);
    }

    /**
     * @param string $pathFilenameSource
     * @param string $pathFilenameThumbnail
     * @param string $dimension
     * @return string
     * @throws UserFormException
     * @throws UserReportException
     */
    private function doThumbnail($pathFilenameSource, $pathFilenameThumbnail, $dimension) {
        $debugMode = false;

        $statSource = stat($pathFilenameSource);
        $statThumbnail = stat($pathFilenameThumbnail);

        if ($statSource === false) {
            throw new UserFormException('File not found: "' . OnString::strrstr($pathFilenameSource, '/') . '"', ERROR_IO_FILE_NOT_FOUND);
        }

        if (isset($statThumbnail['size']) && $statThumbnail['size'] == 0) {

            // If the thumbnail has not been rendered during the last 60 secondes, something is wrong.
            if (time() - $statThumbnail['mtime'] > THUMBNAIL_MAX_SECONDS) {
                $debugMode = true;
//                throw new UserFormException('Thumbnail rendering takes longer than ' . THUMBNAIL_MAX_SECONDS . 's for file: ' .
//                    OnString::strrstr($pathFilenameSource,'/'), ERROR_THUMBNAIL_RENDER);
            } else {
                return $pathFilenameThumbnail; // There seems to be a thumbnail rendering already running. User needs to refresh page.
            }
        }

        if ($statThumbnail === false || $statThumbnail['mtime'] < $statSource['mtime']) {
            $pathFilenameThumbnail = $this->createThumbnail($pathFilenameSource, $pathFilenameThumbnail, $dimension, $debugMode);
        }

        return $pathFilenameThumbnail;
    }

    /**
     * Creates a thumbnail from '$pathFilenameSource' under '$pathFilenameThumbnail'.
     * SVG will be rendered via 'inkscape', all others via 'convert'.
     * Creates an empty '$pathFilenameThumbnail' before the rendering starts. Will be overwritten through the rendering.
     *
     * @param string $pathFilenameSource
     * @param string $pathFilenameThumbnail
     * @param string $dimension
     * @param $debugMode
     * @return string
     * @throws CodeException
     * @throws UserFormException
     * @throws UserReportException
     */
    private function createThumbnail($pathFilenameSource, $pathFilenameThumbnail, $dimension, $debugMode) {
        $outputInkscape = '';
        $cmdInkscape = '';
        $background = $debugMode ? '&' : '';

        // Indicates a running thumbnail rendering process.
        if (false === touch($pathFilenameThumbnail)) {
            // Be sure that the target directory exist
            Support::mkDirParent($pathFilenameThumbnail);
            if (false === touch($pathFilenameThumbnail)) {
                throw new UserReportException('Could not create file: ' . OnString::strrstr($pathFilenameSource, '/'), ERROR_IO_CREATE_FILE);
            }
        }

        // Get extension.
        $ext = strtolower(OnString::strrstr($pathFilenameSource, '.'));

        // SVG files are best to thumbnail via 'inkscape'
        if ($ext == 'svg' && $this->inkscape != '') {
            $inkscapeDimension = Token::explodeDimension($dimension);
            $cmdInkscape = $this->inkscape . " --export-area-drawing --without-gui $inkscapeDimension --export-png $pathFilenameThumbnail $pathFilenameSource $background";
            $outputInkscape = Token::qfqExec($cmdInkscape, $rc);
            if ($rc == 0) {
                return $pathFilenameThumbnail;
            }
            // if process failed, try 'convert' below.
        }

        // On multi page files: just take the first page
        if ($ext == 'pdf') {
            $pathFilenameSource .= '[0]';
        }

        $cmd = $this->convert . " -scale $dimension $pathFilenameSource $pathFilenameThumbnail $background";
        $output = Token::qfqExec($cmd, $rc);

        if ($rc != 0) {
            // Check if the extension is from a supported filetype
            switch ($ext) {

                // The following files should always be thumbnail'able - if not, throw an exception.
                case 'pdf':
                case 'jpg':
                case 'gif':
                case 'png':
                case 'svg':
                    break;

                default:
                    $placeholder = Support::joinPath(THUMBNAIL_UNKNOWN_TYPE, $ext . '.gif');
                    if (is_readable($placeholder)) {
                        return $placeholder;
                    }
            }

            $debug = 'Source: ' . $pathFilenameSource . PHP_EOL;
            $debug .= 'Thumbnail: ' . $pathFilenameThumbnail . PHP_EOL;
            if ($outputInkscape) {
                $debug .= $cmdInkscape . PHP_EOL;
                $debug .= $outputInkscape . PHP_EOL;
            }
            $debug .= $cmd . PHP_EOL;
            $debug .= $output . PHP_EOL;
            $this->store->setVar(SYSTEM_MESSAGE_DEBUG, $debug, STORE_SYSTEM);

            throw new UserReportException("Thumbnail creation failed: " . OnString::strrstr($pathFilenameSource, '/'), ERROR_THUMBNAIL_RENDER);
        }

        return $pathFilenameThumbnail;
    }

    /**
     * @param string $pathFilenameThumbnail
     * @param array $control
     * @return string
     * @internal param bool $flagSecure
     */
    private function buildImageTag($pathFilenameThumbnail, array $control) {

        $src = ($control[TOKEN_SIP] == "1") ? $this->buildDownloadLink($pathFilenameThumbnail) : $pathFilenameThumbnail;

        // With RENDER_MODE_7 return only the URL
        if (isset($control[TOKEN_RENDER]) && $control[TOKEN_RENDER] == RENDER_MODE_7) {
            return $src;
        }

        $attribute = Support::doAttribute('src', $src);

        return "<img $attribute>";

    }

    /**
     * @param string $pathFilenameThumbnail
     * @return string
     * @throws CodeException
     */
    private function buildDownloadLink($pathFilenameThumbnail) {

        $urlParam = API_DIR . '/' . API_DOWNLOAD_PHP . '?' . DOWNLOAD_MODE . '=' . DOWNLOAD_MODE_FILE;
        $urlParam .= '&' . SIP_DOWNLOAD_PARAMETER . '=' . base64_encode(TOKEN_FILE . ':' . $pathFilenameThumbnail);

        $sip = $this->store->getSipInstance();

        return $sip->queryStringToSip($urlParam);
    }
}